"""Distributed, redundant and transactional storage for ZODB
"""

from setuptools import setup, find_packages
import os

classifiers = """\
Framework :: ZODB
Intended Audience :: Developers
License :: OSI Approved :: GNU General Public License (GPL)
Operating System :: POSIX :: Linux
Programming Language :: Python
Topic :: Database
Topic :: Software Development :: Libraries :: Python Modules
"""

if not os.path.exists('mock.py'):
    import cStringIO, md5, urllib, zipfile
    mock_py = zipfile.ZipFile(cStringIO.StringIO(urllib.urlopen(
        'http://downloads.sf.net/sourceforge/python-mock/pythonmock-0.1.0.zip'
        ).read())).read('mock.py')
    if md5.md5(mock_py).hexdigest() != '79f42f390678e5195d9ce4ae43bd18ec':
        raise EnvironmentError("MD5 checksum mismatch downloading 'mock.py'")
    open('mock.py', 'w').write(mock_py)

extras_require = {
    'admin': [],
    'client': ['ZODB3'], # ZODB3 >= 3.10
    'ctl': [],
    'master': [],
    'storage-btree': ['ZODB3'],
    'storage-mysqldb': ['MySQL-python'],
}
extras_require['tests'] = ['zope.testing', 'psutil',
    'neoppod[%s]' % ', '.join(extras_require)]

setup(
    name = 'neoppod',
    version = '0.9.2',
    description = __doc__.strip(),
    author = 'NEOPPOD',
    author_email = 'neo-dev@erp5.org',
    url = 'http://www.neoppod.org/',
    license = 'GPL 2+',
    platforms = ["any"],
    classifiers=classifiers.splitlines(),
    long_description = ".. contents::\n\n" + open('README').read()
                     + "\n" + open('CHANGES').read(),
    packages = find_packages(),
    py_modules = ['mock'],
    entry_points = {
        'console_scripts': [
            # XXX: we'd like not to generate scripts for unwanted features
            # (eg. we don't want neotestrunner if nothing depends on tests)
            'neoadmin=neo.scripts.neoadmin:main',
            'neoctl=neo.scripts.neoctl:main',
            'neomaster=neo.scripts.neomaster:main',
            'neomigrate=neo.scripts.neomigrate:main',
            'neostorage=neo.scripts.neostorage:main',
            'neotestrunner=neo.scripts.runner:main',
            'neosimple=neo.scripts.simple:main',
            'stat_zodb=neo.tests.stat_zodb:main',
        ],
    },
    # Raah!!! I wish I could write something like:
    #  install_requires = ['python>=2.5|ctypes'],
    extras_require = extras_require,
    package_data = {
        'neo.client': [
            'component.xml',
        ],
    },
    zip_safe = True,
)



##############################################################################
#
# Copyright (c) 2011 Nexedi SARL and Contributors. All Rights Reserved.
#                    Julien Muchembled <jm@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
##############################################################################

import math

class CacheItem(object):

    __slots__ = ('oid', 'tid', 'next_tid', 'data',
                 'counter', 'level', 'expire',
                 'prev', 'next')

    def __repr__(self):
        s = ''
        for attr in self.__slots__:
            try:
                value = getattr(self, attr)
                if value:
                    if attr in ('prev', 'next'):
                        s += ' %s=<...>' % attr
                        continue
                    elif attr == 'data':
                        value = '...'
                s += ' %s=%r' % (attr, value)
            except AttributeError:
                pass
        return '<%s%s>' % (self.__class__.__name__, s)

class ClientCache(object):
    """In-memory pickle cache based on Multi-Queue cache algorithm

      Multi-Queue algorithm for Second Level Buffer Caches:
      http://www.usenix.org/event/usenix01/full_papers/zhou/zhou_html/index.html

      Quick description:
      - There are multiple "regular" queues, plus a history queue
      - The queue to store an object in depends on its access frequency
      - The queue an object is in defines its lifespan (higher-index queue eq.
        longer lifespan)
        -> The more often an object is accessed, the higher lifespan it will
           have
      - Upon cache or history hit, object frequency is increased and object
        might get moved to longer-lived queue
      - Each access "ages" objects in cache, and an aging object is moved to
        shorter-lived queue as it ages without being accessed, or in the
        history queue if it's really too old.
    """

    __slots__ = ('_life_time', '_max_history_size', '_max_size',
                 '_queue_list', '_oid_dict', '_time', '_size', '_history_size')

    def __init__(self, life_time=10000, max_history_size=100000,
                                        max_size=20*1024*1024):
        self._life_time = life_time
        self._max_history_size = max_history_size
        self._max_size = max_size
        self.clear()

    def clear(self):
        """Reset cache"""
        self._queue_list = [None] # first is history
        self._oid_dict = {}
        self._time = 0
        self._size = 0
        self._history_size = 0

    def _iterQueue(self, level):
        """for debugging purpose"""
        if level < len(self._queue_list):
            item = head = self._queue_list[level]
            if item:
                while 1:
                    yield item
                    item = item.next
                    if item is head:
                        break

    def _add(self, item):
        level = item.level
        try:
            head = self._queue_list[level]
        except IndexError:
            assert len(self._queue_list) == level
            self._queue_list.append(item)
            item.prev = item.next = item
        else:
            if head:
                item.prev = tail = head.prev
                tail.next = head.prev = item
                item.next = head
            else:
                self._queue_list[level] = item
                item.prev = item.next = item
        if level:
            item.expire = self._time + self._life_time
        else:
            self._size -= len(item.data)
            item.data = None
            if self._history_size < self._max_history_size:
                self._history_size += 1
            else:
                self._remove(head)
                item_list = self._oid_dict[head.oid]
                item_list.remove(head)
                if not item_list:
                    del self._oid_dict[head.oid]

    def _remove(self, item):
        level = item.level
        if level is not None:
            item.level = level - 1
            next = item.next
            if next is item:
                self._queue_list[level] = next = None
            else:
                item.prev.next = next
                next.prev = item.prev
                if self._queue_list[level] is item:
                    self._queue_list[level] = next
            return next

    def _fetched(self, item, _log=math.log):
        self._remove(item)
        item.counter = counter = item.counter + 1
        # XXX It might be better to adjust the level according to the object
        # size. See commented factor for example.
        item.level = 1 + int(_log(counter, 2)
                             # * (1.01 - float(len(item.data)) / self._max_size)
                            )
        self._add(item)

        self._time = time = self._time + 1
        for head in self._queue_list[1:]:
            if head and head.expire < time:
                self._remove(head)
                self._add(head)
                break

    def _load(self, oid, before_tid=None):
        item_list = self._oid_dict.get(oid)
        if item_list:
            if before_tid:
                for item in reversed(item_list):
                    if item.tid < before_tid:
                        next_tid = item.next_tid
                        if next_tid and next_tid < before_tid:
                            break
                        return item
            else:
                item = item_list[-1]
                if not item.next_tid:
                    return item

    def load(self, oid, before_tid=None):
        """Return a revision of oid that was current before given tid"""
        item = self._load(oid, before_tid)
        if item:
            data = item.data
            if data is not None:
                self._fetched(item)
                return data, item.tid, item.next_tid

    def store(self, oid, data, tid, next_tid):
        """Store a new data record in the cache"""
        size = len(data)
        max_size = self._max_size
        if size < max_size:
            item = self._load(oid, next_tid)
            if item:
                assert not (item.data or item.level)
                assert item.tid == tid and item.next_tid == next_tid
                self._history_size -= 1
            else:
                item = CacheItem()
                item.oid = oid
                item.tid = tid
                item.next_tid = next_tid
                item.counter = 0
                item.level = None
                try:
                    item_list = self._oid_dict[oid]
                except KeyError:
                    self._oid_dict[oid] = [item]
                else:
                    if next_tid:
                        for i, x in enumerate(item_list):
                            if tid < x.tid:
                                break
                        item_list.insert(i, item)
                    else:
                        if item_list:
                            prev = item_list[-1]
                            item.counter = prev.counter
                            prev.counter = 0
                            if prev.level > 1:
                                self._fetched(prev)
                        item_list.append(item)
            item.data = data
            self._fetched(item)
            self._size += size
            if max_size < self._size:
                for head in self._queue_list[1:]:
                    while head:
                        next = self._remove(head)
                        head.level = 0
                        self._add(head)
                        if self._size <= max_size:
                            return
                        head = next

    def invalidate(self, oid, tid):
        """Mark data record as being valid only up to given tid"""
        try:
            item = self._oid_dict[oid][-1]
        except KeyError:
            pass
        else:
            if item.next_tid is None:
                item.next_tid = tid



#
# Copyright (C) 2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import traceback
import signal
import ctypes
import imp
import os
import sys
from functools import wraps
import neo

# WARNING: This module should only be used for live application debugging.
# It - by purpose - allows code injection in a running neo process.
# You don't want to enable it in a production environment. Really.
ENABLED = False

# How to include in python code:
#   from neo.debug import register
#   register()
#
# How to trigger it:
#   Kill python process with:
#     SIGUSR1:
#       Loads (or reloads) neo.debug module.
#       The content is up to you (it's only imported).
#     SIGUSR2:
#       Triggers a pdb prompt on process' controlling TTY.

libc = ctypes.cdll.LoadLibrary('libc.so.6')
errno = ctypes.c_int.in_dll(libc, 'errno')

def decorate(func):
    def decorator(sig, frame):
        # Save errno value, to restore it after sig handler returns
        old_errno = errno.value
        try:
            func(sig, frame)
        except:
            # Prevent exception from exiting signal handler, so mistakes in
            # "debug" module don't kill process.
            traceback.print_exc()
        errno.value = old_errno
    return wraps(func)(decorator)

@decorate
def debugHandler(sig, frame):
    file, filename, (suffix, mode, type) = imp.find_module('debug',
        neo.__path__)
    imp.load_module('neo.debug', file, filename, (suffix, mode, type))

def getPdb():
    try: # try ipython if available
        import IPython
        IPython.Shell.IPShell(argv=[])
        return IPython.Debugger.Tracer().debugger
    except ImportError:
        import pdb
        return pdb.Pdb()

_debugger = None

def winpdb(depth=0):
    import rpdb2
    depth += 1
    if rpdb2.g_debugger is not None:
        return rpdb2.setbreak(depth)
    script = rpdb2.calc_frame_path(sys._getframe(depth))
    pwd = str(os.getpid()) + os.getcwd().replace('/', '_').replace('-', '_')
    pid = os.fork()
    if pid:
        try:
            rpdb2.start_embedded_debugger(pwd, depth=depth)
        finally:
            os.waitpid(pid, 0)
    else:
        try:
            os.execlp('python', 'python', '-c', """import os\nif not os.fork():
                import rpdb2, winpdb
                rpdb2_raw_input = rpdb2._raw_input
                rpdb2._raw_input = lambda s: \
                    s == rpdb2.STR_PASSWORD_INPUT and %r or rpdb2_raw_input(s)
                winpdb.g_ignored_warnings[winpdb.STR_EMBEDDED_WARNING] = True
                winpdb.main()
            """ % pwd, '-a', script)
        finally:
            os.abort()

@decorate
def pdbHandler(sig, frame):
    try:
        winpdb(2) # depth is 2, because of the decorator
    except ImportError:
        global _debugger
        if _debugger is None:
            _debugger = getPdb()
        debugger.set_trace(frame)

def register(on_log=None):
    if ENABLED:
        signal.signal(signal.SIGUSR1, debugHandler)
        signal.signal(signal.SIGUSR2, pdbHandler)
        if on_log is not None:
            # use 'kill -RTMIN <pid>
            @decorate
            def on_log_signal(signum, signal):
                on_log()
            signal.signal(signal.SIGRTMIN, on_log_signal)




#
# Copyright (C) 2009-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import __builtin__
import os
import random
import socket
import sys
import tempfile
import unittest
import MySQLdb
import neo
import transaction

from mock import Mock
from neo.lib import debug, logger, protocol, setupLog
from neo.lib.protocol import Packets
from neo.lib.util import getAddressType
from time import time, gmtime
from struct import pack, unpack

DB_PREFIX = os.getenv('NEO_DB_PREFIX', 'test_neo')
DB_ADMIN = os.getenv('NEO_DB_ADMIN', 'root')
DB_PASSWD = os.getenv('NEO_DB_PASSWD', '')
DB_USER = os.getenv('NEO_DB_USER', 'test')

IP_VERSION_FORMAT_DICT = {
    socket.AF_INET:  '127.0.0.1',
    socket.AF_INET6: '::1',
}

ADDRESS_TYPE = socket.AF_INET

debug.ENABLED = True
debug.register()
# prevent "signal only works in main thread" errors in subprocesses
debug.ENABLED = False

def mockDefaultValue(name, function):
    def method(self, *args, **kw):
        if name in self.mockReturnValues:
            return self.__getattr__(name)(*args, **kw)
        return function(self, *args, **kw)
    method.__name__ = name
    setattr(Mock, name, method)

mockDefaultValue('__nonzero__', lambda self: self.__len__() != 0)
mockDefaultValue('__repr__', lambda self:
    '<%s object at 0x%x>' % (self.__class__.__name__, id(self)))
mockDefaultValue('__str__', repr)

def buildUrlFromString(address):
    try:
        socket.inet_pton(socket.AF_INET6, address)
        address = '[%s]' % address
    except Exception:
        pass
    return address

def getTempDirectory():
    """get the current temp directory or a new one"""
    try:
        temp_dir = os.environ['TEMP']
    except KeyError:
        neo_dir = os.path.join(tempfile.gettempdir(), 'neo_tests')
        while True:
            temp_dir = os.path.join(neo_dir, repr(time()))
            try:
                os.makedirs(temp_dir)
                break
            except OSError, e:
                if e.errno != errno.EEXIST:
                    raise
        os.environ['TEMP'] = temp_dir
        print 'Using temp directory %r.' % temp_dir
    return temp_dir

def setupMySQLdb(db_list, user=DB_USER, password='', clear_databases=True):
    from MySQLdb.constants.ER import BAD_DB_ERROR
    conn = MySQLdb.Connect(user=DB_ADMIN, passwd=DB_PASSWD)
    cursor = conn.cursor()
    for database in db_list:
        try:
            conn.select_db(database)
            if not clear_databases:
                continue
            cursor.execute('DROP DATABASE `%s`' % database)
        except MySQLdb.OperationalError, (code, _):
            if code != BAD_DB_ERROR:
                raise
            cursor.execute('GRANT ALL ON `%s`.* TO "%s"@"localhost" IDENTIFIED'
                           ' BY "%s"' % (database, user, password))
        cursor.execute('CREATE DATABASE `%s`' % database)
    cursor.close()
    conn.commit()
    conn.close()

class NeoTestBase(unittest.TestCase):
    def setUp(self):
        logger.PACKET_LOGGER.enable(True)
        sys.stdout.write(' * %s ' % (self.id(), ))
        sys.stdout.flush()
        self.setupLog()
        unittest.TestCase.setUp(self)

    def setupLog(self):
        test_case, test_method = self.id().rsplit('.', 1)
        log_file = os.path.join(getTempDirectory(), test_case + '.log')
        setupLog(test_method, log_file, True)

    def tearDown(self):
        # Kill all unfinished transactions for next test.
        # Note we don't even abort them because it may require a valid
        # connection to a master node (see Storage.sync()).
        transaction.manager.__init__()
        unittest.TestCase.tearDown(self)
        sys.stdout.write('\n')
        sys.stdout.flush()

    failIfEqual = failUnlessEqual = assertEquals = assertNotEquals = None

    def assertNotEqual(self, first, second, msg=None):
        assert not (isinstance(first, Mock) or isinstance(second, Mock)), \
          "Mock objects can't be compared with '==' or '!='"
        return super(NeoTestBase, self).assertNotEqual(first, second, msg=msg)

    def assertEqual(self, first, second, msg=None):
        assert not (isinstance(first, Mock) or isinstance(second, Mock)), \
          "Mock objects can't be compared with '==' or '!='"
        return super(NeoTestBase, self).assertEqual(first, second, msg=msg)

class NeoUnitTestBase(NeoTestBase):
    """ Base class for neo tests, implements common checks """

    local_ip = IP_VERSION_FORMAT_DICT[ADDRESS_TYPE]

    def prepareDatabase(self, number, prefix='test_neo'):
        """ create empties databases """
        setupMySQLdb(['%s%u' % (prefix, i) for i in xrange(number)])

    def getMasterConfiguration(self, cluster='main', master_number=2,
            replicas=2, partitions=1009, uuid=None):
        assert master_number >= 1 and master_number <= 10
        masters = ([(self.local_ip, 10010 + i)
                    for i in xrange(master_number)])
        return Mock({
                'getCluster': cluster,
                'getBind': masters[0],
                'getMasters': (masters, getAddressType((
                        self.local_ip, 0))),
                'getReplicas': replicas,
                'getPartitions': partitions,
                'getUUID': uuid,
        })

    def getStorageConfiguration(self, cluster='main', master_number=2,
            index=0, prefix=DB_PREFIX, uuid=None):
        assert master_number >= 1 and master_number <= 10
        assert index >= 0 and index <= 9
        masters = [(buildUrlFromString(self.local_ip),
                     10010 + i) for i in xrange(master_number)]
        database = '%s@%s%s' % (DB_USER, prefix, index)
        return Mock({
                'getCluster': cluster,
                'getName': 'storage',
                'getBind': (masters[0], 10020 + index),
                'getMasters': (masters, getAddressType((
                        self.local_ip, 0))),
                'getDatabase': database,
                'getUUID': uuid,
                'getReset': False,
                'getAdapter': 'MySQL',
        })

    def _makeUUID(self, prefix):
        """
            Retuns a 16-bytes UUID according to namespace 'prefix'
        """
        assert len(prefix) == 1
        uuid = protocol.INVALID_UUID
        while uuid[1:] == protocol.INVALID_UUID[1:]:
            uuid = prefix + os.urandom(15)
        return uuid

    def getNewUUID(self):
        return self._makeUUID('\0')

    def getClientUUID(self):
        return self._makeUUID('C')

    def getMasterUUID(self):
        return self._makeUUID('M')

    def getStorageUUID(self):
        return self._makeUUID('S')

    def getAdminUUID(self):
        return self._makeUUID('A')

    def getNextTID(self, ltid=None):
        tm = time()
        gmt = gmtime(tm)
        upper = ((((gmt.tm_year - 1900) * 12 + gmt.tm_mon - 1) * 31 \
                  + gmt.tm_mday - 1) * 24 + gmt.tm_hour) * 60 + gmt.tm_min
        lower = int((gmt.tm_sec % 60 + (tm - int(tm))) / (60.0 / 65536.0 / 65536.0))
        tid = pack('!LL', upper, lower)
        if ltid is not None and tid <= ltid:
            upper, lower = unpack('!LL', self._last_tid)
            if lower == 0xffffffff:
                # This should not happen usually.
                from datetime import timedelta, datetime
                d = datetime(gmt.tm_year, gmt.tm_mon, gmt.tm_mday,
                             gmt.tm_hour, gmt.tm_min) \
                        + timedelta(0, 60)
                upper = ((((d.year - 1900) * 12 + d.month - 1) * 31 \
                          + d.day - 1) * 24 + d.hour) * 60 + d.minute
                lower = 0
            else:
                lower += 1
            tid = pack('!LL', upper, lower)
        return tid

    def getPTID(self, i=None):
        """ Return an integer PTID """
        if i is None:
            return random.randint(1, 2**64)
        return i

    def getOID(self, i=None):
        """ Return a 8-bytes OID """
        if i is None:
            return os.urandom(8)
        return pack('!Q', i)

    def getTwoIDs(self):
        """ Return a tuple of two sorted UUIDs """
        # generate two ptid, first is lower
        uuids = self.getNewUUID(), self.getNewUUID()
        return min(uuids), max(uuids)

    def getFakeConnector(self, descriptor=None):
        return Mock({
            '__repr__': 'FakeConnector',
            'getDescriptor': descriptor,
            'getAddress': ('', 0),
        })

    def getFakeConnection(self, uuid=None, address=('127.0.0.1', 10000),
            is_server=False, connector=None, peer_id=None):
        if connector is None:
            connector = self.getFakeConnector()
        return Mock({
            'getUUID': uuid,
            'getAddress': address,
            'isServer': is_server,
            '__repr__': 'FakeConnection',
            '__nonzero__': 0,
            'getConnector': connector,
            'getPeerId': peer_id,
        })

    def checkProtocolErrorRaised(self, method, *args, **kwargs):
        """ Check if the ProtocolError exception was raised """
        self.assertRaises(protocol.ProtocolError, method, *args, **kwargs)

    def checkUnexpectedPacketRaised(self, method, *args, **kwargs):
        """ Check if the UnexpectedPacketError exception wxas raised """
        self.assertRaises(protocol.UnexpectedPacketError, method, *args, **kwargs)

    def checkIdenficationRequired(self, method, *args, **kwargs):
        """ Check is the identification_required decorator is applied """
        self.checkUnexpectedPacketRaised(method, *args, **kwargs)

    def checkBrokenNodeDisallowedErrorRaised(self, method, *args, **kwargs):
        """ Check if the BrokenNodeDisallowedError exception wxas raised """
        self.assertRaises(protocol.BrokenNodeDisallowedError, method, *args, **kwargs)

    def checkNotReadyErrorRaised(self, method, *args, **kwargs):
        """ Check if the NotReadyError exception wxas raised """
        self.assertRaises(protocol.NotReadyError, method, *args, **kwargs)

    def checkAborted(self, conn):
        """ Ensure the connection was aborted """
        self.assertEqual(len(conn.mockGetNamedCalls('abort')), 1)

    def checkNotAborted(self, conn):
        """ Ensure the connection was not aborted """
        self.assertEqual(len(conn.mockGetNamedCalls('abort')), 0)

    def checkClosed(self, conn):
        """ Ensure the connection was closed """
        self.assertEqual(len(conn.mockGetNamedCalls('close')), 1)

    def checkNotClosed(self, conn):
        """ Ensure the connection was not closed """
        self.assertEqual(len(conn.mockGetNamedCalls('close')), 0)

    def _checkNoPacketSend(self, conn, method_id):
        call_list = conn.mockGetNamedCalls(method_id)
        self.assertEqual(len(call_list), 0, call_list)

    def checkNoPacketSent(self, conn, check_notify=True, check_answer=True,
            check_ask=True):
        """ check if no packet were sent """
        if check_notify:
            self._checkNoPacketSend(conn, 'notify')
        if check_answer:
            self._checkNoPacketSend(conn, 'answer')
        if check_ask:
            self._checkNoPacketSend(conn, 'ask')

    def checkNoUUIDSet(self, conn):
        """ ensure no UUID was set on the connection """
        self.assertEqual(len(conn.mockGetNamedCalls('setUUID')), 0)

    def checkUUIDSet(self, conn, uuid=None):
        """ ensure no UUID was set on the connection """
        calls = conn.mockGetNamedCalls('setUUID')
        self.assertEqual(len(calls), 1)
        call = calls.pop()
        if uuid is not None:
            self.assertEqual(call.getParam(0), uuid)

    # in check(Ask|Answer|Notify)Packet we return the packet so it can be used
    # in tests if more accurates checks are required

    def checkErrorPacket(self, conn, decode=False):
        """ Check if an error packet was answered """
        calls = conn.mockGetNamedCalls("answer")
        self.assertEqual(len(calls), 1)
        packet = calls.pop().getParam(0)
        self.assertTrue(isinstance(packet, protocol.Packet))
        self.assertEqual(type(packet), Packets.Error)
        if decode:
            return packet.decode()
            return protocol.decode_table[type(packet)](packet._body)
        return packet

    def checkAskPacket(self, conn, packet_type, decode=False):
        """ Check if an ask-packet with the right type is sent """
        calls = conn.mockGetNamedCalls('ask')
        self.assertEqual(len(calls), 1)
        packet = calls.pop().getParam(0)
        self.assertTrue(isinstance(packet, protocol.Packet))
        self.assertEqual(type(packet), packet_type)
        if decode:
            return packet.decode()
        return packet

    def checkAnswerPacket(self, conn, packet_type, decode=False):
        """ Check if an answer-packet with the right type is sent """
        calls = conn.mockGetNamedCalls('answer')
        self.assertEqual(len(calls), 1)
        packet = calls.pop().getParam(0)
        self.assertTrue(isinstance(packet, protocol.Packet))
        self.assertEqual(type(packet), packet_type)
        if decode:
            return packet.decode()
        return packet

    def checkNotifyPacket(self, conn, packet_type, packet_number=0, decode=False):
        """ Check if a notify-packet with the right type is sent """
        calls = conn.mockGetNamedCalls('notify')
        packet = calls.pop(packet_number).getParam(0)
        self.assertTrue(isinstance(packet, protocol.Packet))
        self.assertEqual(type(packet), packet_type)
        if decode:
            return packet.decode()
        return packet

    def checkNotify(self, conn, **kw):
        return self.checkNotifyPacket(conn, Packets.Notify, **kw)

    def checkNotifyNodeInformation(self, conn, **kw):
        return self.checkNotifyPacket(conn, Packets.NotifyNodeInformation, **kw)

    def checkSendPartitionTable(self, conn, **kw):
        return self.checkNotifyPacket(conn, Packets.SendPartitionTable, **kw)

    def checkStartOperation(self, conn, **kw):
        return self.checkNotifyPacket(conn, Packets.StartOperation, **kw)

    def checkInvalidateObjects(self, conn, **kw):
        return self.checkNotifyPacket(conn, Packets.InvalidateObjects, **kw)

    def checkAbortTransaction(self, conn, **kw):
        return self.checkNotifyPacket(conn, Packets.AbortTransaction, **kw)

    def checkNotifyLastOID(self, conn, **kw):
        return self.checkNotifyPacket(conn, Packets.NotifyLastOID, **kw)

    def checkAnswerTransactionFinished(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerTransactionFinished, **kw)

    def checkAnswerInformationLocked(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerInformationLocked, **kw)

    def checkAskLockInformation(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskLockInformation, **kw)

    def checkNotifyUnlockInformation(self, conn, **kw):
        return self.checkNotifyPacket(conn, Packets.NotifyUnlockInformation, **kw)

    def checkNotifyTransactionFinished(self, conn, **kw):
        return self.checkNotifyPacket(conn, Packets.NotifyTransactionFinished, **kw)

    def checkRequestIdentification(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.RequestIdentification, **kw)

    def checkAskPrimary(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskPrimary)

    def checkAskUnfinishedTransactions(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskUnfinishedTransactions)

    def checkAskTransactionInformation(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskTransactionInformation, **kw)

    def checkAskObjectPresent(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskObjectPresent, **kw)

    def checkAskObject(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskObject, **kw)

    def checkAskStoreObject(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskStoreObject, **kw)

    def checkAskStoreTransaction(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskStoreTransaction, **kw)

    def checkAskFinishTransaction(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskFinishTransaction, **kw)

    def checkAskNewTid(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskBeginTransaction, **kw)

    def checkAskLastIDs(self, conn, **kw):
        return self.checkAskPacket(conn, Packets.AskLastIDs, **kw)

    def checkAcceptIdentification(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AcceptIdentification, **kw)

    def checkAnswerPrimary(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerPrimary, **kw)

    def checkAnswerLastIDs(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerLastIDs, **kw)

    def checkAnswerUnfinishedTransactions(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerUnfinishedTransactions, **kw)

    def checkAnswerObject(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerObject, **kw)

    def checkAnswerTransactionInformation(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerTransactionInformation, **kw)

    def checkAnswerBeginTransaction(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerBeginTransaction, **kw)

    def checkAnswerTids(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerTIDs, **kw)

    def checkAnswerTidsFrom(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerTIDsFrom, **kw)

    def checkAnswerObjectHistory(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerObjectHistory, **kw)

    def checkAnswerObjectHistoryFrom(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerObjectHistoryFrom, **kw)

    def checkAnswerStoreTransaction(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerStoreTransaction, **kw)

    def checkAnswerStoreObject(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerStoreObject, **kw)

    def checkAnswerOids(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerOIDs, **kw)

    def checkAnswerPartitionTable(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerPartitionTable, **kw)

    def checkAnswerObjectPresent(self, conn, **kw):
        return self.checkAnswerPacket(conn, Packets.AnswerObjectPresent, **kw)

connector_cpt = 0

class DoNothingConnector(Mock):
    def __init__(self, s=None):
        neo.lib.logging.info("initializing connector")
        global connector_cpt
        self.desc = connector_cpt
        connector_cpt += 1
        self.packet_cpt = 0
        Mock.__init__(self)

    def getAddress(self):
        return self.addr

    def makeClientConnection(self, addr):
        self.addr = addr

    def makeListeningConnection(self, addr):
        self.addr = addr

    def getDescriptor(self):
        return self.desc


__builtin__.pdb = lambda depth=0: \
    debug.getPdb().set_trace(sys._getframe(depth+1))



#
# Copyright (C) 2006-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import neo.lib
from neo.lib.protocol import NodeStates, ErrorCodes, Packets, Errors
from neo.lib.protocol import PacketMalformedError, UnexpectedPacketError, \
        BrokenNodeDisallowedError, NotReadyError, ProtocolError


class EventHandler(object):
    """This class handles events."""

    def __init__(self, app):
        self.app = app

    def __repr__(self):
        return self.__class__.__name__

    def __unexpectedPacket(self, conn, packet, message=None):
        """Handle an unexpected packet."""
        if message is None:
            message = 'unexpected packet type %s in %s' % (type(packet),
                    self.__class__.__name__)
        else:
            message = 'unexpected packet: %s in %s' % (message,
                    self.__class__.__name__)
        neo.lib.logging.error(message)
        conn.answer(Errors.ProtocolError(message))
        conn.abort()
        # self.peerBroken(conn)

    def dispatch(self, conn, packet):
        """This is a helper method to handle various packet types."""
        try:
            try:
                method = getattr(self, packet.handler_method_name)
            except AttributeError:
                raise UnexpectedPacketError('no handler found')
            args = packet.decode() or ()
            conn.setPeerId(packet.getId())
            method(conn, *args)
        except UnexpectedPacketError, e:
            self.__unexpectedPacket(conn, packet, *e.args)
        except PacketMalformedError:
            neo.lib.logging.error('malformed packet from %r', conn)
            conn.notify(Packets.Notify('Malformed packet: %r' % (packet, )))
            conn.abort()
            # self.peerBroken(conn)
        except BrokenNodeDisallowedError:
            conn.answer(Errors.BrokenNode('go away'))
            conn.abort()
        except NotReadyError, message:
            if not message.args:
                message = 'Retry Later'
            message = str(message)
            conn.answer(Errors.NotReady(message))
            conn.abort()
        except ProtocolError, message:
            message = str(message)
            conn.answer(Errors.ProtocolError(message))
            conn.abort()

    def checkClusterName(self, name):
        # raise an exception if the given name mismatch the current cluster name
        if self.app.name != name:
            neo.lib.logging.error('reject an alien cluster')
            raise ProtocolError('invalid cluster name')


    # Network level handlers

    def packetReceived(self, conn, packet):
        """Called when a packet is received."""
        self.dispatch(conn, packet)

    def connectionStarted(self, conn):
        """Called when a connection is started."""
        neo.lib.logging.debug('connection started for %r', conn)

    def connectionCompleted(self, conn):
        """Called when a connection is completed."""
        neo.lib.logging.debug('connection completed for %r (from %s:%u)',
                              conn, *conn.getConnector().getAddress())

    def connectionFailed(self, conn):
        """Called when a connection failed."""
        neo.lib.logging.debug('connection failed for %r', conn)

    def connectionAccepted(self, conn):
        """Called when a connection is accepted."""

    def connectionClosed(self, conn):
        """Called when a connection is closed by the peer."""
        neo.lib.logging.debug('connection closed for %r', conn)
        self.connectionLost(conn, NodeStates.TEMPORARILY_DOWN)

    #def peerBroken(self, conn):
    #    """Called when a peer is broken."""
    #    neo.lib.logging.error('%r is broken', conn)
    #    # NodeStates.BROKEN

    def connectionLost(self, conn, new_state):
        """ this is a method to override in sub-handlers when there is no need
        to make distinction from the kind event that closed the connection  """
        pass


    # Packet handlers.

    def ping(self, conn):
        if not conn.isAborted():
            conn.answer(Packets.Pong())

    def pong(self, conn):
        # Ignore PONG packets. The only purpose of ping/pong packets is
        # to test/maintain underlying connection.
        pass

    def notify(self, conn, message):
        neo.lib.logging.info('notification from %r: %s', conn, message)

    def askBarrier(self, conn):
        conn.answer(Packets.AnswerBarrier())

    def answerBarrier(self, conn):
        pass

    # Error packet handlers.

    def error(self, conn, code, message):
        try:
            getattr(self, Errors[code])(conn, message)
        except (AttributeError, ValueError):
            raise UnexpectedPacketError(message)

    def protocolError(self, conn, message):
        # the connection should have been closed by the remote peer
        neo.lib.logging.error('protocol error: %s' % (message,))

    def timeoutError(self, conn, message):
        neo.lib.logging.error('timeout error: %s' % (message,))

    def brokenNodeDisallowedError(self, conn, message):
        raise RuntimeError, 'broken node disallowed error: %s' % (message,)

    def alreadyPendingError(self, conn, message):
        neo.lib.logging.error('already pending error: %s' % (message, ))

    def ack(self, conn, message):
        neo.lib.logging.debug("no error message : %s" % (message))



#
# Copyright (C) 2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import unittest
from mock import Mock
from neo.lib.util import add64
from neo.tests import NeoUnitTestBase
from neo.lib.protocol import Packets, ZERO_OID, ZERO_TID
from neo.storage.handlers.replication import ReplicationHandler
from neo.storage.handlers.replication import RANGE_LENGTH, MIN_RANGE_LENGTH

class FakeDict(object):
    def __init__(self, items):
        self._items = items
        self._dict = dict(items)
        assert len(self._dict) == len(items), self._dict

    def iteritems(self):
        for item in self._items:
            yield item

    def iterkeys(self):
        for key, value in self.iteritems():
            yield key

    def itervalues(self):
        for key, value in self.iteritems():
            yield value

    def items(self):
        return self._items[:]

    def keys(self):
        return [x for x, y in self._items]

    def values(self):
        return [y for x, y in self._items]

    def __getitem__(self, key):
        return self._dict[key]

    def __getattr__(self, key):
        return getattr(self._dict, key)

    def __len__(self):
        return len(self._dict)

class StorageReplicationHandlerTests(NeoUnitTestBase):

    def setup(self):
        pass

    def teardown(self):
        pass

    def getApp(self, conn=None, tid_check_result=(0, 0, ZERO_TID),
                serial_check_result=(0, 0, ZERO_OID, 0, ZERO_TID),
                tid_result=(),
                history_result=None,
                rid=0, critical_tid=ZERO_TID,
                num_partitions=1,
            ):
        if history_result is None:
            history_result = {}
        replicator = Mock({
            '__repr__': 'Fake replicator',
            'reset': None,
            'checkSerialRange': None,
            'checkTIDRange': None,
            'getTIDCheckResult': tid_check_result,
            'getSerialCheckResult': serial_check_result,
            'getTIDsFromResult': tid_result,
            'getObjectHistoryFromResult': history_result,
            'checkSerialRange': None,
            'checkTIDRange': None,
            'getTIDsFrom': None,
            'getObjectHistoryFrom': None,
            'getCurrentOffset': rid,
            'getCurrentCriticalTID': critical_tid,
        })
        def isCurrentConnection(other_conn):
            return other_conn is conn
        replicator.isCurrentConnection = isCurrentConnection
        real_replicator = replicator
        class FakeApp(object):
            replicator = real_replicator
            dm = Mock({
                'storeTransaction': None,
                'deleteObject': None,
            })
            pt = Mock({
                'getPartitions': num_partitions,
            })
        return FakeApp

    def _checkReplicationStarted(self, conn, rid, replicator):
        min_tid, max_tid, length, partition = self.checkAskPacket(conn,
            Packets.AskCheckTIDRange, decode=True)
        self.assertEqual(min_tid, ZERO_TID)
        self.assertEqual(length, RANGE_LENGTH)
        self.assertEqual(partition, rid)
        calls = replicator.mockGetNamedCalls('checkTIDRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(min_tid, max_tid, length, partition)

    def _checkPacketTIDList(self, conn, tid_list, next_tid, app):
        packet_list = [x.getParam(0) for x in conn.mockGetNamedCalls('ask')]
        packet_list, next_range = packet_list[:-1], packet_list[-1]

        self.assertEqual(type(next_range), Packets.AskCheckTIDRange)
        pmin_tid, plength, ppartition = next_range.decode()
        self.assertEqual(pmin_tid, add64(next_tid, 1))
        self.assertEqual(plength, RANGE_LENGTH)
        self.assertEqual(ppartition, app.replicator.getCurrentOffset())
        calls = app.replicator.mockGetNamedCalls('checkTIDRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_tid, plength, ppartition)

        self.assertEqual(len(packet_list), len(tid_list))
        for packet in packet_list:
            self.assertEqual(type(packet),
                Packets.AskTransactionInformation)
            ptid = packet.decode()[0]
            for tid in tid_list:
                if ptid == tid:
                    tid_list.remove(tid)
                    break
            else:
                raise AssertionFailed, '%s not found in %r' % (dump(ptid),
                    [dump(x) for x in tid_list])

    def _checkPacketSerialList(self, conn, object_list, next_oid, next_serial, app):
        packet_list = [x.getParam(0) for x in conn.mockGetNamedCalls('ask')]
        packet_list, next_range = packet_list[:-1], packet_list[-1]

        self.assertEqual(type(next_range), Packets.AskCheckSerialRange)
        pmin_oid, pmin_serial, plength, ppartition = next_range.decode()
        self.assertEqual(pmin_oid, next_oid)
        self.assertEqual(pmin_serial, add64(next_serial, 1))
        self.assertEqual(plength, RANGE_LENGTH)
        self.assertEqual(ppartition, app.replicator.getCurrentOffset())
        calls = app.replicator.mockGetNamedCalls('checkSerialRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_oid, pmin_serial, plength, ppartition)

        self.assertEqual(len(packet_list), len(object_list),
            ([x.decode() for x in packet_list], object_list))
        reference_set = set((x + (None, ) for x in object_list))
        packet_set = set((x.decode() for x in packet_list))
        assert len(packet_list) == len(reference_set) == len(packet_set)
        self.assertEqual(reference_set, packet_set)

    def test_connectionLost(self):
        app = self.getApp()
        ReplicationHandler(app).connectionLost(None, None)
        self.assertEqual(len(app.replicator.mockGetNamedCalls('storageLost')), 1)

    def test_connectionFailed(self):
        app = self.getApp()
        ReplicationHandler(app).connectionFailed(None)
        self.assertEqual(len(app.replicator.mockGetNamedCalls('storageLost')), 1)

    def test_acceptIdentification(self):
        rid = 24
        app = self.getApp(rid=rid)
        conn = self.getFakeConnection()
        replication = ReplicationHandler(app)
        replication.acceptIdentification(conn, None, None, None,
            None, None)
        self._checkReplicationStarted(conn, rid, app.replicator)

    def test_startReplication(self):
        rid = 24
        app = self.getApp(rid=rid)
        conn = self.getFakeConnection()
        ReplicationHandler(app).startReplication(conn)
        self._checkReplicationStarted(conn, rid, app.replicator)

    def test_answerTIDsFrom(self):
        conn = self.getFakeConnection()
        tid_list = [self.getOID(0), self.getOID(1), self.getOID(2)]
        app = self.getApp(conn=conn, tid_result=[])
        # With no known TID
        ReplicationHandler(app).answerTIDsFrom(conn, tid_list)
        # With some TIDs known
        conn = self.getFakeConnection()
        known_tid_list = [tid_list[0], tid_list[1]]
        unknown_tid_list = [tid_list[2], ]
        app = self.getApp(conn=conn, tid_result=known_tid_list)
        ReplicationHandler(app).answerTIDsFrom(conn, tid_list[1:])
        calls = app.dm.mockGetNamedCalls('deleteTransaction')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(tid_list[0])

    def test_answerTransactionInformation(self):
        conn = self.getFakeConnection()
        app = self.getApp(conn=conn)
        tid = self.getNextTID()
        user = 'foo'
        desc = 'bar'
        ext = 'baz'
        packed = True
        oid_list = [self.getOID(1), self.getOID(2)]
        ReplicationHandler(app).answerTransactionInformation(conn, tid, user,
            desc, ext, packed, oid_list)
        calls = app.dm.mockGetNamedCalls('storeTransaction')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(tid, (), (oid_list, user, desc, ext, packed), False)

    def test_answerObjectHistoryFrom(self):
        conn = self.getFakeConnection()
        oid_1 = self.getOID(1)
        oid_2 = self.getOID(2)
        oid_3 = self.getOID(3)
        oid_4 = self.getOID(4)
        oid_5 = self.getOID(5)
        tid_list = [self.getOID(x) for x in xrange(7)]
        oid_dict = FakeDict((
            (oid_1, [tid_list[0], tid_list[1]]),
            (oid_2, [tid_list[2], tid_list[3]]),
            (oid_4, [tid_list[5]]),
        ))
        flat_oid_list = []
        for oid, serial_list in oid_dict.iteritems():
            for serial in serial_list:
                flat_oid_list.append((oid, serial))
        app = self.getApp(conn=conn, history_result={})
        # With no known OID/Serial
        ReplicationHandler(app).answerObjectHistoryFrom(conn, oid_dict)
        # With some known OID/Serials
        # For test to be realist, history_result should contain the same
        # number of serials as oid_dict, otherise it just tests the special
        # case of the last check in a partition.
        conn = self.getFakeConnection()
        app = self.getApp(conn=conn, history_result={
            oid_1: [oid_dict[oid_1][0], ],
            oid_3: [tid_list[2]],
            oid_4: [tid_list[4], oid_dict[oid_4][0], tid_list[6]],
            oid_5: [tid_list[6]],
        })
        ReplicationHandler(app).answerObjectHistoryFrom(conn, oid_dict)
        calls = app.dm.mockGetNamedCalls('deleteObject')
        actual_deletes = set(((x.getParam(0), x.getParam(1)) for x in calls))
        expected_deletes = set((
            (oid_3, tid_list[2]),
            (oid_4, tid_list[4]),
        ))
        self.assertEqual(actual_deletes, expected_deletes)

    def test_answerObject(self):
        conn = self.getFakeConnection()
        app = self.getApp(conn=conn)
        oid = self.getOID(1)
        serial_start = self.getNextTID()
        serial_end = self.getNextTID()
        compression = 1
        checksum = 2
        data = 'foo'
        data_serial = None
        ReplicationHandler(app).answerObject(conn, oid, serial_start,
            serial_end, compression, checksum, data, data_serial)
        calls = app.dm.mockGetNamedCalls('storeTransaction')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(serial_start, [(oid, compression, checksum, data,
            data_serial)], None, False)

    # CheckTIDRange
    def test_answerCheckTIDFullRangeIdenticalChunkWithNext(self):
        min_tid = self.getNextTID()
        max_tid = self.getNextTID()
        critical_tid = self.getNextTID()
        assert max_tid < critical_tid
        length = RANGE_LENGTH
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length, 0, max_tid), rid=rid,
            conn=conn, critical_tid=critical_tid)
        handler = ReplicationHandler(app)
        # Peer has the same data as we have: length, checksum and max_tid
        # match.
        handler.answerCheckTIDRange(conn, min_tid, length, length, 0, max_tid)
        # Result: go on with next chunk
        pmin_tid, pmax_tid, plength, ppartition = self.checkAskPacket(conn,
            Packets.AskCheckTIDRange, decode=True)
        self.assertEqual(pmin_tid, add64(max_tid, 1))
        self.assertEqual(plength, RANGE_LENGTH)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('checkTIDRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_tid, pmax_tid, plength, ppartition)

    def test_answerCheckTIDSmallRangeIdenticalChunkWithNext(self):
        min_tid = self.getNextTID()
        max_tid = self.getNextTID()
        critical_tid = self.getNextTID()
        assert max_tid < critical_tid
        length = RANGE_LENGTH / 2
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length, 0, max_tid), rid=rid,
            conn=conn, critical_tid=critical_tid)
        handler = ReplicationHandler(app)
        # Peer has the same data as we have: length, checksum and max_tid
        # match.
        handler.answerCheckTIDRange(conn, min_tid, length, length, 0, max_tid)
        # Result: go on with next chunk
        pmin_tid, pmax_tid, plength, ppartition = self.checkAskPacket(conn,
            Packets.AskCheckTIDRange, decode=True)
        self.assertEqual(pmax_tid, critical_tid)
        self.assertEqual(pmin_tid, add64(max_tid, 1))
        self.assertEqual(plength, length / 2)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('checkTIDRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_tid, pmax_tid, plength, ppartition)

    def test_answerCheckTIDRangeIdenticalChunkAboveCriticalTID(self):
        critical_tid = self.getNextTID()
        min_tid = self.getNextTID()
        max_tid = self.getNextTID()
        assert critical_tid < max_tid
        length = RANGE_LENGTH / 2
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length, 0, max_tid), rid=rid,
            conn=conn, critical_tid=critical_tid)
        handler = ReplicationHandler(app)
        # Peer has the same data as we have: length, checksum and max_tid
        # match.
        handler.answerCheckTIDRange(conn, min_tid, length, length, 0, max_tid)
        # Result: go on with object range checks
        pmin_oid, pmin_serial, pmax_tid, plength, ppartition = \
            self.checkAskPacket(conn, Packets.AskCheckSerialRange, decode=True)
        self.assertEqual(pmin_oid, ZERO_OID)
        self.assertEqual(pmin_serial, ZERO_TID)
        self.assertEqual(plength, RANGE_LENGTH)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('checkSerialRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_oid, pmin_serial, pmax_tid, plength, ppartition)

    def test_answerCheckTIDRangeIdenticalChunkWithoutNext(self):
        min_tid = self.getNextTID()
        max_tid = self.getNextTID()
        length = RANGE_LENGTH / 2
        rid = 12
        num_partitions = 13
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length - 1, 0, max_tid), rid=rid,
            conn=conn, num_partitions=num_partitions)
        handler = ReplicationHandler(app)
        # Peer has the same data as we have: length, checksum and max_tid
        # match.
        handler.answerCheckTIDRange(conn, min_tid, length, length - 1, 0,
            max_tid)
        # Result: go on with object range checks
        pmin_oid, pmin_serial, pmax_tid, plength, ppartition = \
            self.checkAskPacket(conn, Packets.AskCheckSerialRange, decode=True)
        self.assertEqual(pmin_oid, ZERO_OID)
        self.assertEqual(pmin_serial, ZERO_TID)
        self.assertEqual(plength, RANGE_LENGTH)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('checkSerialRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_oid, pmin_serial, pmax_tid, plength, ppartition)
        # ...and delete partition tail
        calls = app.dm.mockGetNamedCalls('deleteTransactionsAbove')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(num_partitions, rid, add64(max_tid, 1), ZERO_TID)

    def test_answerCheckTIDRangeDifferentBigChunk(self):
        min_tid = self.getNextTID()
        max_tid = self.getNextTID()
        critical_tid = self.getNextTID()
        assert min_tid < max_tid < critical_tid, (min_tid, max_tid,
            critical_tid)
        length = RANGE_LENGTH / 2
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length - 5, 0, max_tid), rid=rid,
            conn=conn, critical_tid=critical_tid)
        handler = ReplicationHandler(app)
        # Peer has different data
        handler.answerCheckTIDRange(conn, min_tid, length, length, 0, max_tid)
        # Result: ask again, length halved
        pmin_tid, pmax_tid, plength, ppartition = self.checkAskPacket(conn,
            Packets.AskCheckTIDRange, decode=True)
        self.assertEqual(pmin_tid, min_tid)
        self.assertEqual(plength, length / 2)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('checkTIDRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_tid, pmax_tid, plength, ppartition)

    def test_answerCheckTIDRangeDifferentSmallChunkWithNext(self):
        min_tid = self.getNextTID()
        max_tid = self.getNextTID()
        critical_tid = self.getNextTID()
        length = MIN_RANGE_LENGTH - 1
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length - 5, 0, max_tid), rid=rid,
            conn=conn, critical_tid=critical_tid)
        handler = ReplicationHandler(app)
        # Peer has different data
        handler.answerCheckTIDRange(conn, min_tid, length, length, 0, max_tid)
        # Result: ask tid list, and ask next chunk
        calls = conn.mockGetNamedCalls('ask')
        self.assertEqual(len(calls), 1)
        tid_packet = calls[0].getParam(0)
        self.assertEqual(type(tid_packet), Packets.AskTIDsFrom)
        pmin_tid, pmax_tid, plength, ppartition = tid_packet.decode()
        self.assertEqual(pmin_tid, min_tid)
        self.assertEqual(pmax_tid, critical_tid)
        self.assertEqual(plength, length)
        self.assertEqual(ppartition, [rid])
        calls = app.replicator.mockGetNamedCalls('getTIDsFrom')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_tid, pmax_tid, plength, ppartition[0])

    def test_answerCheckTIDRangeDifferentSmallChunkWithoutNext(self):
        min_tid = self.getNextTID()
        max_tid = self.getNextTID()
        critical_tid = self.getNextTID()
        length = MIN_RANGE_LENGTH - 1
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length - 5, 0, max_tid), rid=rid,
            conn=conn, critical_tid=critical_tid)
        handler = ReplicationHandler(app)
        # Peer has different data, and less than length
        handler.answerCheckTIDRange(conn, min_tid, length, length - 1, 0,
            max_tid)
        # Result: ask tid list, and start replicating object range
        calls = conn.mockGetNamedCalls('ask')
        self.assertEqual(len(calls), 2)
        tid_packet = calls[0].getParam(0)
        self.assertEqual(type(tid_packet), Packets.AskTIDsFrom)
        pmin_tid, pmax_tid, plength, ppartition = tid_packet.decode()
        self.assertEqual(pmin_tid, min_tid)
        self.assertEqual(pmax_tid, critical_tid)
        self.assertEqual(plength, length - 1)
        self.assertEqual(ppartition, [rid])
        calls = app.replicator.mockGetNamedCalls('getTIDsFrom')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_tid, pmax_tid, plength, ppartition[0])

    # CheckSerialRange
    def test_answerCheckSerialFullRangeIdenticalChunkWithNext(self):
        min_oid = self.getOID(1)
        max_oid = self.getOID(10)
        min_serial = self.getNextTID()
        max_serial = self.getNextTID()
        length = RANGE_LENGTH
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(serial_check_result=(length, 0, max_oid, 1,
            max_serial), rid=rid, conn=conn)
        handler = ReplicationHandler(app)
        # Peer has the same data as we have
        handler.answerCheckSerialRange(conn, min_oid, min_serial, length,
            length, 0, max_oid, 1, max_serial)
        # Result: go on with next chunk
        pmin_oid, pmin_serial, pmax_tid, plength, ppartition = \
            self.checkAskPacket(conn, Packets.AskCheckSerialRange, decode=True)
        self.assertEqual(pmin_oid, max_oid)
        self.assertEqual(pmin_serial, add64(max_serial, 1))
        self.assertEqual(plength, RANGE_LENGTH)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('checkSerialRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_oid, pmin_serial, pmax_tid, plength, ppartition)

    def test_answerCheckSerialSmallRangeIdenticalChunkWithNext(self):
        min_oid = self.getOID(1)
        max_oid = self.getOID(10)
        min_serial = self.getNextTID()
        max_serial = self.getNextTID()
        length = RANGE_LENGTH / 2
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(serial_check_result=(length, 0, max_oid, 1,
            max_serial), rid=rid, conn=conn)
        handler = ReplicationHandler(app)
        # Peer has the same data as we have
        handler.answerCheckSerialRange(conn, min_oid, min_serial, length,
            length, 0, max_oid, 1, max_serial)
        # Result: go on with next chunk
        pmin_oid, pmin_serial, pmax_tid, plength, ppartition = \
            self.checkAskPacket(conn, Packets.AskCheckSerialRange, decode=True)
        self.assertEqual(pmin_oid, max_oid)
        self.assertEqual(pmin_serial, add64(max_serial, 1))
        self.assertEqual(plength, length / 2)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('checkSerialRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_oid, pmin_serial, pmax_tid, plength, ppartition)

    def test_answerCheckSerialRangeIdenticalChunkWithoutNext(self):
        min_oid = self.getOID(1)
        max_oid = self.getOID(10)
        min_serial = self.getNextTID()
        max_serial = self.getNextTID()
        length = RANGE_LENGTH / 2
        rid = 12
        num_partitions = 13
        conn = self.getFakeConnection()
        app = self.getApp(serial_check_result=(length - 1, 0, max_oid, 1,
            max_serial), rid=rid, conn=conn, num_partitions=num_partitions)
        handler = ReplicationHandler(app)
        # Peer has the same data as we have
        handler.answerCheckSerialRange(conn, min_oid, min_serial, length,
            length - 1, 0, max_oid, 1, max_serial)
        # Result: mark replication as done
        self.checkNoPacketSent(conn)
        self.assertTrue(app.replicator.replication_done)
        # ...and delete partition tail
        calls = app.dm.mockGetNamedCalls('deleteObjectsAbove')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(num_partitions, rid, max_oid, add64(max_serial, 1),
            ZERO_TID)

    def test_answerCheckSerialRangeDifferentBigChunk(self):
        min_oid = self.getOID(1)
        max_oid = self.getOID(10)
        min_serial = self.getNextTID()
        max_serial = self.getNextTID()
        length = RANGE_LENGTH / 2
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length - 5, 0, max_oid, 1,
            max_serial), rid=rid, conn=conn)
        handler = ReplicationHandler(app)
        # Peer has different data
        handler.answerCheckSerialRange(conn, min_oid, min_serial, length,
            length, 0, max_oid, 1, max_serial)
        # Result: ask again, length halved
        pmin_oid, pmin_serial, pmax_tid, plength, ppartition = \
            self.checkAskPacket(conn, Packets.AskCheckSerialRange, decode=True)
        self.assertEqual(pmin_oid, min_oid)
        self.assertEqual(pmin_serial, min_serial)
        self.assertEqual(plength, length / 2)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('checkSerialRange')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_oid, pmin_serial, pmax_tid, plength, ppartition)

    def test_answerCheckSerialRangeDifferentSmallChunkWithNext(self):
        min_oid = self.getOID(1)
        max_oid = self.getOID(10)
        min_serial = self.getNextTID()
        max_serial = self.getNextTID()
        critical_tid = self.getNextTID()
        length = MIN_RANGE_LENGTH - 1
        rid = 12
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length - 5, 0, max_oid, 1,
            max_serial), rid=rid, conn=conn, critical_tid=critical_tid)
        handler = ReplicationHandler(app)
        # Peer has different data
        handler.answerCheckSerialRange(conn, min_oid, min_serial, length,
            length, 0, max_oid, 1, max_serial)
        # Result: ask serial list, and ask next chunk
        calls = conn.mockGetNamedCalls('ask')
        self.assertEqual(len(calls), 1)
        serial_packet = calls[0].getParam(0)
        self.assertEqual(type(serial_packet), Packets.AskObjectHistoryFrom)
        pmin_oid, pmin_serial, pmax_serial, plength, ppartition = \
            serial_packet.decode()
        self.assertEqual(pmin_oid, min_oid)
        self.assertEqual(pmin_serial, min_serial)
        self.assertEqual(pmax_serial, critical_tid)
        self.assertEqual(plength, length)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('getObjectHistoryFrom')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_oid, pmin_serial, pmax_serial, plength,
            ppartition)

    def test_answerCheckSerialRangeDifferentSmallChunkWithoutNext(self):
        min_oid = self.getOID(1)
        max_oid = self.getOID(10)
        min_serial = self.getNextTID()
        max_serial = self.getNextTID()
        critical_tid = self.getNextTID()
        length = MIN_RANGE_LENGTH - 1
        rid = 12
        num_partitions = 13
        conn = self.getFakeConnection()
        app = self.getApp(tid_check_result=(length - 5, 0, max_oid,
            1, max_serial), rid=rid, conn=conn, critical_tid=critical_tid,
            num_partitions=num_partitions,
        )
        handler = ReplicationHandler(app)
        # Peer has different data, and less than length
        handler.answerCheckSerialRange(conn, min_oid, min_serial, length,
            length - 1, 0, max_oid, 1, max_serial)
        # Result: ask tid list, and mark replication as done
        pmin_oid, pmin_serial, pmax_serial, plength, ppartition = \
            self.checkAskPacket(conn, Packets.AskObjectHistoryFrom,
            decode=True)
        self.assertEqual(pmin_oid, min_oid)
        self.assertEqual(pmin_serial, min_serial)
        self.assertEqual(pmax_serial, critical_tid)
        self.assertEqual(plength, length - 1)
        self.assertEqual(ppartition, rid)
        calls = app.replicator.mockGetNamedCalls('getObjectHistoryFrom')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(pmin_oid, pmin_serial, pmax_serial, plength,
            ppartition)
        self.assertTrue(app.replicator.replication_done)
        # ...and delete partition tail
        calls = app.dm.mockGetNamedCalls('deleteObjectsAbove')
        self.assertEqual(len(calls), 1)
        calls[0].checkArgs(num_partitions, rid, max_oid, add64(max_serial, 1),
            critical_tid)

if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2009-2010  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import unittest
from mock import Mock
from neo.tests import NeoUnitTestBase
from neo.lib.protocol import NodeStates, CellStates
from neo.master.pt import PartitionTable
from neo.lib.node import StorageNode

class MasterPartitionTableTests(NeoUnitTestBase):

    def test_02_PartitionTable_creation(self):
        num_partitions = 5
        num_replicas = 3
        pt = PartitionTable(num_partitions, num_replicas)
        self.assertEqual(pt.np, num_partitions)
        self.assertEqual(pt.nr, num_replicas)
        self.assertEqual(pt.num_filled_rows, 0)
        partition_list = pt.partition_list
        self.assertEqual(len(partition_list), num_partitions)
        for x in xrange(num_partitions):
            part = partition_list[x]
            self.assertTrue(isinstance(part, list))
            self.assertEqual(len(part), 0)
        self.assertEqual(len(pt.count_dict), 0)
        # no nodes or cells for now
        self.assertEqual(len(pt.getNodeList()), 0)
        for x in xrange(num_partitions):
            self.assertEqual(len(pt.getCellList(x)), 0)
            self.assertEqual(len(pt.getCellList(x, True)), 0)
            self.assertEqual(len(pt.getRow(x)), 0)
        self.assertFalse(pt.operational())
        self.assertFalse(pt.filled())
        self.assertRaises(RuntimeError, pt.make, [])
        self.assertFalse(pt.operational())
        self.assertFalse(pt.filled())

    def test_11_findLeastUsedNode(self):
        num_partitions = 5
        num_replicas = 2
        pt = PartitionTable(num_partitions, num_replicas)
        # add nodes
        uuid1 = self.getNewUUID()
        server1 = ("127.0.0.1", 19001)
        sn1 = StorageNode(Mock(), server1, uuid1, NodeStates.RUNNING)
        pt.setCell(0, sn1, CellStates.UP_TO_DATE)
        pt.setCell(1, sn1, CellStates.UP_TO_DATE)
        pt.setCell(2, sn1, CellStates.UP_TO_DATE)
        uuid2 = self.getNewUUID()
        server2 = ("127.0.0.2", 19001)
        sn2 = StorageNode(Mock(), server2, uuid2, NodeStates.RUNNING)
        pt.setCell(0, sn2, CellStates.UP_TO_DATE)
        pt.setCell(1, sn2, CellStates.UP_TO_DATE)
        uuid3 = self.getNewUUID()
        server3 = ("127.0.0.3", 19001)
        sn3 = StorageNode(Mock(), server3, uuid3, NodeStates.RUNNING)
        pt.setCell(0, sn3, CellStates.UP_TO_DATE)
        # test
        node = pt.findLeastUsedNode()
        self.assertEqual(node, sn3)
        node = pt.findLeastUsedNode((sn3, ))
        self.assertEqual(node, sn2)
        node = pt.findLeastUsedNode((sn3, sn2))
        self.assertEqual(node, sn1)

    def test_13_outdate(self):
        # create nodes
        uuid1 = self.getNewUUID()
        server1 = ("127.0.0.1", 19001)
        sn1 = StorageNode(Mock(), server1, uuid1)
        uuid2 = self.getNewUUID()
        server2 = ("127.0.0.2", 19002)
        sn2 = StorageNode(Mock(), server2, uuid2)
        uuid3 = self.getNewUUID()
        server3 = ("127.0.0.3", 19003)
        sn3 = StorageNode(Mock(), server3, uuid3)
        uuid4 = self.getNewUUID()
        server4 = ("127.0.0.4", 19004)
        sn4 = StorageNode(Mock(), server4, uuid4)
        uuid5 = self.getNewUUID()
        server5 = ("127.0.0.5", 19005)
        sn5 = StorageNode(Mock(), server5, uuid5)
        # create partition table
        num_partitions = 5
        num_replicas = 3
        pt = PartitionTable(num_partitions, num_replicas)
        pt.setCell(0, sn1, CellStates.OUT_OF_DATE)
        sn1.setState(NodeStates.RUNNING)
        pt.setCell(1, sn2, CellStates.UP_TO_DATE)
        sn2.setState(NodeStates.TEMPORARILY_DOWN)
        pt.setCell(2, sn3, CellStates.UP_TO_DATE)
        sn3.setState(NodeStates.DOWN)
        pt.setCell(3, sn4, CellStates.UP_TO_DATE)
        sn4.setState(NodeStates.BROKEN)
        pt.setCell(4, sn5, CellStates.UP_TO_DATE)
        sn5.setState(NodeStates.RUNNING)
        # outdate nodes
        cells_outdated = pt.outdate()
        self.assertEqual(len(cells_outdated), 3)
        for offset, uuid, state in cells_outdated:
            self.assertTrue(offset in (1, 2, 3))
            self.assertTrue(uuid in (uuid2, uuid3, uuid4))
            self.assertEqual(state, CellStates.OUT_OF_DATE)
        # check each cell
        # part 1, already outdated
        cells = pt.getCellList(0)
        self.assertEqual(len(cells), 1)
        cell = cells[0]
        self.assertEqual(cell.getState(), CellStates.OUT_OF_DATE)
        # part 2, must be outdated
        cells = pt.getCellList(1)
        self.assertEqual(len(cells), 1)
        cell = cells[0]
        self.assertEqual(cell.getState(), CellStates.OUT_OF_DATE)
        # part 3, must be outdated
        cells = pt.getCellList(2)
        self.assertEqual(len(cells), 1)
        cell = cells[0]
        self.assertEqual(cell.getState(), CellStates.OUT_OF_DATE)
        # part 4, already outdated
        cells = pt.getCellList(3)
        self.assertEqual(len(cells), 1)
        cell = cells[0]
        self.assertEqual(cell.getState(), CellStates.OUT_OF_DATE)
        # part 5, remains running
        cells = pt.getCellList(4)
        self.assertEqual(len(cells), 1)
        cell = cells[0]
        self.assertEqual(cell.getState(), CellStates.UP_TO_DATE)

    def test_14_addNode(self):
        num_partitions = 5
        num_replicas = 2
        pt = PartitionTable(num_partitions, num_replicas)
        # add nodes
        uuid1 = self.getNewUUID()
        server1 = ("127.0.0.1", 19001)
        sn1 = StorageNode(Mock(), server1, uuid1)
        # add it to an empty pt
        cell_list = pt.addNode(sn1)
        self.assertEqual(len(cell_list), 5)
        # it must be added to all partitions
        for x in xrange(num_replicas):
            self.assertEqual(len(pt.getCellList(x)), 1)
            self.assertEqual(pt.getCellList(x)[0].getState(), CellStates.OUT_OF_DATE)
            self.assertEqual(pt.getCellList(x)[0].getNode(), sn1)
        self.assertEqual(pt.count_dict[sn1], 5)
        # add same node again, must remain the same
        cell_list = pt.addNode(sn1)
        self.assertEqual(len(cell_list), 0)
        for x in xrange(num_replicas):
            self.assertEqual(len(pt.getCellList(x)), 1)
            self.assertEqual(pt.getCellList(x)[0].getState(), CellStates.OUT_OF_DATE)
            self.assertEqual(pt.getCellList(x)[0].getNode(), sn1)
        self.assertEqual(pt.count_dict[sn1], 5)
        # add a second node to fill the partition table
        uuid2 = self.getNewUUID()
        server2 = ("127.0.0.2", 19002)
        sn2 = StorageNode(Mock(), server2, uuid2)
        # add it
        cell_list = pt.addNode(sn2)
        self.assertEqual(len(cell_list), 5)
        for x in xrange(num_replicas):
            self.assertEqual(len(pt.getCellList(x)), 2)
            self.assertEqual(pt.getCellList(x)[0].getState(), CellStates.OUT_OF_DATE)
            self.assertTrue(pt.getCellList(x)[0].getNode() in (sn1, sn2))
        # test the most used node is remove from some partition
        uuid3 = self.getNewUUID()
        server3 = ("127.0.0.3", 19001)
        sn3 = StorageNode(Mock(), server3, uuid3)
        uuid4 = self.getNewUUID()
        server4 = ("127.0.0.4", 19001)
        sn4 = StorageNode(Mock(), server4, uuid4)
        uuid5 = self.getNewUUID()
        server5 = ("127.0.0.5", 1900)
        sn5 = StorageNode(Mock(), server5, uuid5)
        # partition looks like:
        # 0 : sn1, sn2
        # 1 : sn1, sn3
        # 2 : sn1, sn4
        # 3 : sn1, sn5
        num_partitions = 4
        num_replicas = 1
        pt = PartitionTable(num_partitions, num_replicas)
        # node most used is out of date, just dropped
        pt.setCell(0, sn1, CellStates.OUT_OF_DATE)
        pt.setCell(0, sn2, CellStates.UP_TO_DATE)
        pt.setCell(1, sn1, CellStates.OUT_OF_DATE)
        pt.setCell(1, sn3, CellStates.UP_TO_DATE)
        pt.setCell(2, sn1, CellStates.OUT_OF_DATE)
        pt.setCell(2, sn4, CellStates.UP_TO_DATE)
        pt.setCell(3, sn1, CellStates.OUT_OF_DATE)
        pt.setCell(3, sn5, CellStates.UP_TO_DATE)
        uuid6 = self.getNewUUID()
        server6 = ("127.0.0.6", 19006)
        sn6 = StorageNode(Mock(), server6, uuid6)
        cell_list = pt.addNode(sn6)
        # sn1 is removed twice and sn6 is added twice
        self.assertEqual(len(cell_list), 4)
        for offset, uuid, state  in cell_list:
            if offset in (0, 1):
                if uuid == uuid1:
                    self.assertEqual(state, CellStates.DISCARDED)
                elif uuid == uuid6:
                    self.assertEqual(state, CellStates.OUT_OF_DATE)
                else:
                    self.assertTrue(uuid in (uuid1, uuid6))
            else:
                self.assertTrue(offset in (0, 1))
        for x in xrange(num_replicas):
            self.assertEqual(len(pt.getCellList(x)), 2)
        # there is a feeding cell, just dropped
        pt.clear()
        pt.setCell(0, sn1, CellStates.UP_TO_DATE)
        pt.setCell(0, sn2, CellStates.UP_TO_DATE)
        pt.setCell(0, sn3, CellStates.FEEDING)
        pt.setCell(1, sn1, CellStates.UP_TO_DATE)
        pt.setCell(1, sn2, CellStates.FEEDING)
        pt.setCell(1, sn3, CellStates.UP_TO_DATE)
        pt.setCell(2, sn1, CellStates.UP_TO_DATE)
        pt.setCell(2, sn4, CellStates.FEEDING)
        pt.setCell(2, sn5, CellStates.UP_TO_DATE)
        pt.setCell(3, sn1, CellStates.UP_TO_DATE)
        pt.setCell(3, sn4, CellStates.UP_TO_DATE)
        pt.setCell(3, sn5, CellStates.FEEDING)
        cell_list = pt.addNode(sn6)
        # sn1 is removed twice and sn6 is added twice
        self.assertEqual(len(cell_list), 4)
        for offset, uuid, state  in cell_list:
            if offset in (0, 1):
                if uuid == uuid1:
                    self.assertEqual(state, CellStates.DISCARDED)
                elif uuid == uuid6:
                    self.assertEqual(state, CellStates.OUT_OF_DATE)
                else:
                    self.assertTrue(uuid in (uuid1, uuid6))
            else:
                self.assertTrue(offset in (0, 1))
        for x in xrange(num_replicas):
            self.assertEqual(len(pt.getCellList(x)), 3)
        # there is no feeding cell, marked as feeding
        pt.clear()
        pt.setCell(0, sn1, CellStates.UP_TO_DATE)
        pt.setCell(0, sn2, CellStates.UP_TO_DATE)
        pt.setCell(1, sn1, CellStates.UP_TO_DATE)
        pt.setCell(1, sn3, CellStates.UP_TO_DATE)
        pt.setCell(2, sn1, CellStates.UP_TO_DATE)
        pt.setCell(2, sn4, CellStates.UP_TO_DATE)
        pt.setCell(3, sn1, CellStates.UP_TO_DATE)
        pt.setCell(3, sn5, CellStates.UP_TO_DATE)
        cell_list = pt.addNode(sn6)
        # sn1 is removed twice and sn6 is added twice
        self.assertEqual(len(cell_list), 4)
        for offset, uuid, state  in cell_list:
            if offset in (0, 1):
                if uuid == uuid1:
                    self.assertEqual(state, CellStates.FEEDING)
                elif uuid == uuid6:
                    self.assertEqual(state, CellStates.OUT_OF_DATE)
                else:
                    self.assertTrue(uuid in (uuid1, uuid6))
            else:
                self.assertTrue(offset in (0, 1))
        for x in xrange(num_replicas):
            self.assertEqual(len(pt.getCellList(x)), 3)

    def test_15_dropNode(self):
        num_partitions = 4
        num_replicas = 2
        pt = PartitionTable(num_partitions, num_replicas)
        # add nodes
        uuid1 = self.getNewUUID()
        server1 = ("127.0.0.1", 19001)
        sn1 = StorageNode(Mock(), server1, uuid1, NodeStates.RUNNING)
        uuid2 = self.getNewUUID()
        server2 = ("127.0.0.2", 19002)
        sn2 = StorageNode(Mock(), server2, uuid2, NodeStates.RUNNING)
        uuid3 = self.getNewUUID()
        server3 = ("127.0.0.3", 19001)
        sn3 = StorageNode(Mock(), server3, uuid3, NodeStates.RUNNING)
        uuid4 = self.getNewUUID()
        server4 = ("127.0.0.4", 19001)
        sn4 = StorageNode(Mock(), server4, uuid4, NodeStates.RUNNING)
        # partition looks like:
        # 0 : sn1, sn2
        # 1 : sn1, sn3
        # 2 : sn1, sn3
        # 3 : sn1, sn4
        # node is not feeding, so retrive least use node to replace it
        # so sn2 must be repaced by sn4 in partition 0
        pt.setCell(0, sn1, CellStates.UP_TO_DATE)
        pt.setCell(0, sn2, CellStates.UP_TO_DATE)
        pt.setCell(1, sn1, CellStates.UP_TO_DATE)
        pt.setCell(1, sn3, CellStates.UP_TO_DATE)
        pt.setCell(2, sn1, CellStates.UP_TO_DATE)
        pt.setCell(2, sn3, CellStates.UP_TO_DATE)
        pt.setCell(3, sn1, CellStates.UP_TO_DATE)
        pt.setCell(3, sn4, CellStates.UP_TO_DATE)
        cell_list = pt.dropNode(sn2)
        self.assertEqual(len(cell_list), 2)
        for offset, uuid, state  in cell_list:
            self.assertEqual(offset, 0)
            if uuid == uuid2:
                self.assertEqual(state, CellStates.DISCARDED)
            elif uuid == uuid4:
                self.assertEqual(state, CellStates.OUT_OF_DATE)
            else:
                self.assertTrue(uuid in (uuid2, uuid4))

        for x in xrange(num_replicas):
            self.assertEqual(len(pt.getCellList(x)), 2)
        # same test but with feeding state, no other will be added
        pt.clear()
        pt.setCell(0, sn1, CellStates.UP_TO_DATE)
        pt.setCell(0, sn2, CellStates.FEEDING)
        pt.setCell(1, sn1, CellStates.UP_TO_DATE)
        pt.setCell(1, sn3, CellStates.UP_TO_DATE)
        pt.setCell(2, sn1, CellStates.UP_TO_DATE)
        pt.setCell(2, sn3, CellStates.UP_TO_DATE)
        pt.setCell(3, sn1, CellStates.UP_TO_DATE)
        pt.setCell(3, sn4, CellStates.UP_TO_DATE)
        cell_list = pt.dropNode(sn2)
        self.assertEqual(len(cell_list), 1)
        for offset, uuid, state  in cell_list:
            self.assertEqual(offset, 0)
            self.assertEqual(state, CellStates.DISCARDED)
            self.assertEqual(uuid, uuid2)
        for x in xrange(num_replicas):
            if x == 0:
                self.assertEqual(len(pt.getCellList(x)), 1)
            else:
                self.assertEqual(len(pt.getCellList(x)), 2)

    def test_16_make(self):
        num_partitions = 5
        num_replicas = 1
        pt = PartitionTable(num_partitions, num_replicas)
        # add nodes
        uuid1 = self.getNewUUID()
        server1 = ("127.0.0.1", 19001)
        sn1 = StorageNode(Mock(), server1, uuid1, NodeStates.RUNNING)
        # add not running node
        uuid2 = self.getNewUUID()
        server2 = ("127.0.0.2", 19001)
        sn2 = StorageNode(Mock(), server2, uuid2)
        sn2.setState(NodeStates.TEMPORARILY_DOWN)
        # add node without uuid
        server3 = ("127.0.0.3", 19001)
        sn3 = StorageNode(Mock(), server3, None, NodeStates.RUNNING)
        # add clear node
        uuid4 = self.getNewUUID()
        server4 = ("127.0.0.4", 19001)
        sn4 = StorageNode(Mock(), server4, uuid4, NodeStates.RUNNING)
        uuid5 = self.getNewUUID()
        server5 = ("127.0.0.5", 1900)
        sn5 = StorageNode(Mock(), server5, uuid5, NodeStates.RUNNING)
        # make the table
        pt.make([sn1, sn2, sn3, sn4, sn5])
        # check it's ok, only running nodes and node with uuid
        # must be present
        for x in xrange(num_partitions):
            cells = pt.getCellList(x)
            self.assertEqual(len(cells), 2)
            nodes = [x.getNode() for x in cells]
            for node in nodes:
                self.assertTrue(node in (sn1, sn4, sn5))
                self.assertTrue(node not in (sn2, sn3))
        self.assertTrue(pt.filled())
        self.assertTrue(pt.operational())
        # create a pt with less nodes
        pt.clear()
        self.assertFalse(pt.filled())
        self.assertFalse(pt.operational())
        pt.make([sn1])
        # check it's ok
        for x in xrange(num_partitions):
            cells = pt.getCellList(x)
            self.assertEqual(len(cells), 1)
            nodes = [x.getNode() for x in cells]
            for node in nodes:
                self.assertEqual(node, sn1)
        self.assertTrue(pt.filled())
        self.assertTrue(pt.operational())

    def test_17_tweak(self):
        # remove broken node
        # remove if too many feeding nodes
        # remove feeding if all cells are up to date
        # if too many cells, remove most used cell
        # if not enought cell, add least used node

        # create nodes
        uuid1 = self.getNewUUID()
        server1 = ("127.0.0.1", 19001)
        sn1 = StorageNode(Mock(), server1, uuid1, NodeStates.RUNNING)
        uuid2 = self.getNewUUID()
        server2 = ("127.0.0.2", 19002)
        sn2 = StorageNode(Mock(), server2, uuid2, NodeStates.RUNNING)
        uuid3 = self.getNewUUID()
        server3 = ("127.0.0.3", 19003)
        sn3 = StorageNode(Mock(), server3, uuid3, NodeStates.RUNNING)
        uuid4 = self.getNewUUID()
        server4 = ("127.0.0.4", 19004)
        sn4 = StorageNode(Mock(), server4, uuid4, NodeStates.RUNNING)
        uuid5 = self.getNewUUID()
        server5 = ("127.0.0.5", 19005)
        sn5 = StorageNode(Mock(), server5, uuid5, NodeStates.RUNNING)
        # create partition table
        # 0 : sn1(discarded), sn2(up), -> sn2 must remain
        # 1 : sn1(feeding), sn2(feeding), sn3(up) -> one feeding and sn3 must remain
        # 2 : sn1(feeding), sn2(up), sn3(up) -> sn2 and sn3 must remain, feeding must go away
        # 3 : sn1(up), sn2(up), sn3(up), sn4(up) -> only 3 cell must remain
        # 4 : sn1(up), sn5(up) -> one more cell must be added
        num_partitions = 5
        num_replicas = 2
        pt = PartitionTable(num_partitions, num_replicas)
        # part 0
        pt.setCell(0, sn1, CellStates.DISCARDED)
        pt.setCell(0, sn2, CellStates.UP_TO_DATE)
        # part 1
        pt.setCell(1, sn1, CellStates.FEEDING)
        pt.setCell(1, sn2, CellStates.FEEDING)
        pt.setCell(1, sn3, CellStates.OUT_OF_DATE)
        # part 2
        pt.setCell(2, sn1, CellStates.FEEDING)
        pt.setCell(2, sn2, CellStates.UP_TO_DATE)
        pt.setCell(2, sn3, CellStates.UP_TO_DATE)
        # part 3
        pt.setCell(3, sn1, CellStates.UP_TO_DATE)
        pt.setCell(3, sn2, CellStates.UP_TO_DATE)
        pt.setCell(3, sn3, CellStates.UP_TO_DATE)
        pt.setCell(3, sn4, CellStates.UP_TO_DATE)
        # part 4
        pt.setCell(4, sn1, CellStates.UP_TO_DATE)
        pt.setCell(4, sn5, CellStates.UP_TO_DATE)
        # now tweak the table
        pt.tweak()
        # check part 1
        cells =  pt.getCellList(0)
        self.assertEqual(len(cells), 3)
        for cell in cells:
            self.assertNotEqual(cell.getState(), CellStates.DISCARDED)
            if cell.getNode() == sn2:
                self.assertEqual(cell.getState(), CellStates.UP_TO_DATE)
            else:
                self.assertEqual(cell.getState(), CellStates.OUT_OF_DATE)
        self.assertTrue(sn2 in [x.getNode() for x in cells])

        # check part 2
        cells =  pt.getCellList(1)
        self.assertEqual(len(cells), 4)
        for cell in cells:
            if cell.getNode() == sn1:
                self.assertEqual(cell.getState(), CellStates.FEEDING)
            else:
                self.assertEqual(cell.getState(), CellStates.OUT_OF_DATE)
        self.assertTrue(sn3 in [x.getNode() for x in cells])
        self.assertTrue(sn1 in [x.getNode() for x in cells])

        # check part 3
        cells =  pt.getCellList(2)
        self.assertEqual(len(cells), 3)
        for cell in cells:
            if cell.getNode() in (sn2, sn3):
                self.assertEqual(cell.getState(), CellStates.UP_TO_DATE)
            else:
                self.assertEqual(cell.getState(), CellStates.OUT_OF_DATE)
        self.assertTrue(sn3 in [x.getNode() for x in cells])
        self.assertTrue(sn2 in [x.getNode() for x in cells])

        # check part 4
        cells =  pt.getCellList(3)
        self.assertEqual(len(cells), 3)
        for cell in cells:
            self.assertEqual(cell.getState(), CellStates.UP_TO_DATE)

        # check part 5
        cells =  pt.getCellList(4)
        self.assertEqual(len(cells), 3)
        for cell in cells:
            if cell.getNode() in (sn1, sn5):
                self.assertEqual(cell.getState(), CellStates.UP_TO_DATE)
            else:
                self.assertEqual(cell.getState(), CellStates.OUT_OF_DATE)
        self.assertTrue(sn1 in [x.getNode() for x in cells])
        self.assertTrue(sn5 in [x.getNode() for x in cells])


if __name__ == '__main__':
    unittest.main()

